#!/usr/bin/python
# Copyright 2016 CERN and GSI
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import fesa.fillFesaDesignBase

import os
import sys
import time
import zlib
import glob
import re
import datetime
import socket

import iecommon
import fesaGeneral
import fesaTemplates
import iefiles

from iecommon import *
from fesaGeneral import *

import libxml2


class FESADesignGenerator3_0_0(object):

    #=========================================================================
    # Sub-functions
    #=========================================================================
        
    def isGSITemplate(self,parent):
        if isChildElement(parent,"GSI-Status-Property"):
            return True
        if isChildElement(parent,"GSI-Init-Property"):
            return True
        return False


    def getSilecsStringLength(self,register):
        strlen = '64'# check if custom length exists for string, otherwise use default length (64)
        if(register.hasProp('string-len')):
            strlen = str(register.prop('string-len'))
        return strlen

    def getOrCreateStringArrayType(self,fieldNode,register):
        array2DNode = getOrCreateChildElement(fieldNode,'array2D')
        fillAttributes(array2DNode, {'type': iecommon.getFesaDataType(register.prop('format'))})
        dim1Node = getOrCreateChildElement(array2DNode,'dim1')
        dim2Node = getOrCreateChildElement(array2DNode,'dim2')
        dim1Node.setContent(register.prop('array-dim1')) # first dimension is the dimension of the array
        dim2Node.setContent(self.getSilecsStringLength(register)) # first dimension is the dimension of the array
        return array2DNode

    def getOrCreateStringType(self,fieldNode,register):
        stringNode = getOrCreateChildElement(fieldNode,'array')
        fillAttributes(stringNode, {'type': iecommon.getFesaDataType(register.prop('format'))})
        dimNode = getOrCreateChildElement(stringNode,'dim')
        dimNode.setContent(self.getSilecsStringLength(register)) # first dimension is the dimension of the array
        return stringNode

    def getOrCreate2DArrayType(self,fieldNode,register):
        array2DNode = getOrCreateChildElement(fieldNode,'array2D')
        fillAttributes(array2DNode, {'type': iecommon.getFesaDataType(register.prop('format'))})
        dim1Node = getOrCreateChildElement(array2DNode,'dim1')
        dim2Node = getOrCreateChildElement(array2DNode,'dim2')
        dim1Node.setContent(register.prop('array-dim1'))
        dim2Node.setContent(register.prop('array-dim2'))
        return array2DNode

    def getOrCreateArrayType(self,fieldNode,register):
        arrayNode = getOrCreateChildElement(fieldNode,'array')
        fillAttributes(arrayNode, {'type': iecommon.getFesaDataType(register.prop('format'))})
        dimNode = getOrCreateChildElement(arrayNode,'dim')
        dimNode.setContent(register.prop('array-dim1'))
        return arrayNode

    def getOrCreateScalarType(self,fieldNode,scalarType):
        scalarNode = getOrCreateChildElement(fieldNode,'scalar')
        fillAttributes(scalarNode, {'type': iecommon.getFesaDataType(scalarType)})
        return scalarNode

    def getOrCreateStringTypeBase(self,fieldNode,register):
        if(register.prop('array-dim2')) and int(register.prop('array-dim2')) > 1:
            iecommon.logError('ERROR: In register '+register.prop('name')+' - 2D array of strings not supported in FESA.', True, {'errorlog': True})
            return null
        elif (register.prop('array-dim1')) and int(register.prop('array-dim1')) > 1:
            return self.getOrCreateStringArrayType(fieldNode,register)
        else:
            return self.getOrCreateStringType(fieldNode,register)

    def getOrCreateBaseType(self,fieldNode,register):
        if (register.prop('array-dim2')) and int(register.prop('array-dim2')) > 1:
            return self.getOrCreate2DArrayType(fieldNode,register)
        elif (register.prop('array-dim1')) and int(register.prop('array-dim1')) > 1:
            return self.getOrCreateArrayType(fieldNode,register)
        else:
            return self.getOrCreateScalarType(fieldNode,register.prop('format'))

    def getOrCreateType(self,fieldNode,register):
        if(register.prop('format') == 'string'):
            return self.getOrCreateStringTypeBase(fieldNode,register)
        else:
            return self.getOrCreateBaseType(fieldNode,register)

    def getOrCreateFieldRef(self,valueItemNode,FieldNameRef):
        fieldRefNode = getOrCreateNamedChildElement(valueItemNode,'data-field-ref',FieldNameRef,'field-name-ref')
        return fieldRefNode
        
    def getOrCreateAcqStampItem(self,fieldNode):
        item = getOrCreateChildElement(fieldNode,'acq-stamp-item')
        fillAttributes(item, {'direction': 'OUT', 'name': 'acqStamp'})
        scalar = getOrCreateChildElement(item,'scalar')
        fillAttributes(scalar, {'type': 'int64_t'})
        return item

    def getOrCreateUpdateFlagItem(self,fieldNode):
        item = getOrCreateChildElement(fieldNode,'update-flag-item')
        fillAttributes(item, {'direction': 'OUT', 'name': 'updateFlags', 'optional': 'true'})
        scalar = getOrCreateChildElement(item,'builtin-type-scalar')
        fillAttributes(scalar, {'data-type-name-ref': 'NOTIFICATION_UPDATE'})
        return item

    def getOrCreateCyleNameItem(self,fieldNode):
        item = getOrCreateChildElement(fieldNode,'cycle-name-item')
        fillAttributes(item, {'direction': 'OUT', 'name': 'cycleName', 'optional': 'true'})
        array = getOrCreateChildElement(item,'array')
        fillAttributes(array, {'type': 'char'})
        dim = getOrCreateChildElement(array,'dim')
        dim.setContent('32')
        return item

    def getOrCreateCyleStampItem(self,fieldNode):
        item = getOrCreateChildElement(fieldNode,'cycle-stamp-item')
        fillAttributes(item, {'direction': 'OUT', 'name': 'cycleStamp', 'optional': 'true'})
        scalar = getOrCreateChildElement(item,'scalar')
        fillAttributes(scalar, {'type': 'int64_t'})
        return item

    def getOrCreateAcquisitionContextItem(self,propertyNode):
        item = propertyNode.xpathEval('acquisition-context-item')
        if item != None:
            return item
            
        item = getOrCreateChildElement(propertyNode,'acquisition-context-item')
        item.setProp("direction","OUT")
        acqStamp = getOrCreateChildElement(item,'acqStamp')
        acqStamp.setProp("direction","OUT")
        acqStamp.setProp("name","acqStampGSI")
        acqStampScalar = getOrCreateChildElement(acqStamp,'scalar')
        acqStampScalar.setProp('type','int64_t')
        
        cycleStamp = getOrCreateChildElement(item,'cycleStamp')
        cycleStamp.setProp("direction","OUT")
        cycleStamp.setProp("name","cycleStampGSI")
        cycleStampScalar = getOrCreateChildElement(cycleStamp,'scalar')
        cycleStampScalar.setProp('type','int64_t')
        
        cycleName = getOrCreateChildElement(item,'cycleName')
        cycleName.setProp("direction","OUT")
        cycleName.setProp("name","cycleNameGSI")
        cycleNameArray = getOrCreateChildElement(cycleName,'array')
        cycleNameArray.setProp('type','char')
        cycleNameDim = getOrCreateChildElement(cycleNameArray,'custom-constant-dim')
        cycleNameDim.setProp('constant-name-ref', 'MAX_CYCLE_NAME_LENGTH')
        
        beamProcessID = getOrCreateChildElement(item,'beamProcessID')
        beamProcessID.setProp("direction","OUT")
        beamProcessID.setProp("name","beamProcessID")
        beamProcessIDScalar = getOrCreateChildElement(beamProcessID,'scalar')
        beamProcessIDScalar.setProp('type','int32_t')
        
        sequenceID = getOrCreateChildElement(item,'sequenceID')
        sequenceID.setProp("direction","OUT")
        sequenceID.setProp("name","sequenceID")
        sequenceIDScalar = getOrCreateChildElement(sequenceID,'scalar')
        sequenceIDScalar.setProp('type','int32_t')
        
        fieldRef = getOrCreateChildElement(item,'acquisition-context-field-ref')
        fieldRef.setProp("field-name-ref","acquisitionContext")
                        
    def getOrCreateAction(self,propNode,name,type,actionsNode,implementation):
        iecommon.logDebug('Generating Server-Action: ' + name, {'debuglog': True})
        action = getOrCreateChildElement(propNode,type +'-action')
        if not isChildElement(action,'server-action-ref'):
            actionSub = getOrCreateChildElement(action,'server-action-ref')
            fillAttributes(actionSub, {'server-action-name-ref': name})
        finalName = action.xpathEval("server-action-ref")[0].prop('server-action-name-ref') # could be different from name if already exists
        getActionNode = getOrCreateNamedChildElement(actionsNode,type + '-server-action',finalName)
        fillAttributes(getActionNode, {'implementation': implementation})

    def getOrCreateFieldNode(self,parent,register): #in order to place them before any GSI-specifc fields (order matters!)
        iecommon.logDebug('Generating Data-Field for register: ' + register.prop('name'), {'debuglog': True})
        if( hasChildren(parent) ):
            return getOrCreateNamedPreviousSiblingElement(parent,getFirstChild(parent),'field',getFesaName(register))
        else:
            return getOrCreateNamedChildElement(parent,'field',getFesaName(register))

    def getOrCreatePLCHostNameField(self,configurationNode):
        fieldNode = getOrCreateNamedFirstChild(configurationNode,'field','plcHostName')
        descriptionNode = getOrCreateChildElement(fieldNode,'description')
        descriptionNode.setContent('Hostname of the PLC that contains the related SILECS class device')
        array = getOrCreateChildElement(fieldNode,'array')
        fillAttributes(array, {'type': 'char'})
        dim = getOrCreateChildElement(array,'dim')
        dim.setContent('128')
        return fieldNode

    def getOrCreatePLCDeviceLabelField(self,configurationNode):
        fieldNode = getOrCreateNamedFirstChild(configurationNode,'field','plcDeviceLabel')
        descriptionNode = getOrCreateChildElement(fieldNode,'description')
        descriptionNode.setContent('Name of the related SILECS instance within the PLC mapping')
        array = getOrCreateChildElement(fieldNode,'array')
        fillAttributes(array, {'type': 'char'})
        dim = getOrCreateChildElement(array,'dim')
        dim.setContent('128')
        return fieldNode
        
    def getOrCreateParameterFileField(self,configurationNode):
        fieldNode = getOrCreateNamedFirstChild(configurationNode,'field','parameterFile')
        descriptionNode = getOrCreateChildElement(fieldNode,'description')
        descriptionNode.setContent('ParameterFile of the PLC (*.silecsparam)')
        array = getOrCreateChildElement(fieldNode,'array')
        fillAttributes(array, {'type': 'char'})
        dim = getOrCreateChildElement(array,'dim')
        defaultNode = getOrCreateChildElement(fieldNode,'default')
        defaultNode.setContent('../../../generated/client/MyControllerName.silecsparam')
        dim.setContent('512')
        return fieldNode

    def getOrCreatePLCClassVersionField(self,configurationNode,plcClassVersion):
        fieldNode = getOrCreateNamedFirstChild(configurationNode,'field','plcClassVersion')
        descriptionNode = getOrCreateChildElement(fieldNode,'description')
        descriptionNode.setContent('Version of the SILECS class that needs to be deployed in the controller')
        array = getOrCreateChildElement(fieldNode,'array')
        fillAttributes(array, {'type': 'char'})
        dim = getOrCreateChildElement(array,'dim')
        defaultNode = getOrCreateChildElement(fieldNode,'default')
        defaultNode.setContent(plcClassVersion)
        dim.setContent(str(len(defaultNode.getContent())))
        return fieldNode
                
    def getOrCreateRegisterValueItems(self,prop,block,isSetting):
        for register in block.xpathEval('Register'):
            if register.prop('generateFesaValueItem') == 'true':
                self.getOrCreateValueItem(prop,register,isSetting)

    def getOrCreateSettingProperty(self,parent,block):
        iecommon.logDebug('Generating SettingProperty for Block: ' + block.prop('name'), {'debuglog': True})
        if( hasChildren(parent)):
            propsAfterCommand = parent.xpathEval("*[not(name()='command-property')]")
            prop = getOrCreateNamedPreviousSiblingElement(parent,propsAfterCommand[0], 'setting-property',getFesaName(block))
        else:
            prop = getOrCreateNamedChildElement(parent,'setting-property',getFesaName(block))
        self.getOrCreateRegisterValueItems(prop,block,True)
        self.getOrCreateUpdateFlagItem(prop)
        self.getOrCreateCyleNameItem(prop)
        return prop
            
    def getOrCreateAcquisitionProperty(self,parent,actionsNode,block):
        iecommon.logDebug('Generating AcquisitionProperty for Block: ' + block.prop('name'), {'debuglog': True})
        prop = ""
        if( hasChildren(parent)):
            prop = getOrCreateNamedPreviousSiblingElement(parent,getFirstChild(parent), 'acquisition-property',getFesaName(block))
        else:
            prop = getOrCreateNamedChildElement(parent,'acquisition-property',getFesaName(block))
        fillAttributes(prop, {'visibility': 'development', 'subscribable': 'true', 'multiplexed': 'false', 'on-change': 'true'})
        self.getOrCreateRegisterValueItems(prop,block,False)
        self.getOrCreateAcqStampItem(prop)
        self.getOrCreateUpdateFlagItem(prop)
        self.getOrCreateCyleNameItem(prop)
        self.getOrCreateCyleStampItem(prop)
        self.getOrCreateAction(prop,'Get'+getFesaName(block),'get',actionsNode,'default')
        return prop
                
            
    def getOrCreateGSISettingProperty(self,parent,block):
        iecommon.logDebug('Generating GSISettingProperty for Block: ' + block.prop('name'), {'debuglog': True})
        powerProp = parent.xpathEval('GSI-Power-Property')[0] # template is used --> there has to be a power prop
        prop =  getOrCreateNamedPreviousSiblingElement(parent,powerProp, 'GSI-Setting-Property',getFesaName(block))
        self.getOrCreateRegisterValueItems(prop,block,True)
        self.getOrCreateUpdateFlagItem(prop)
        self.getOrCreateCyleNameItem(prop)
        return prop
            
    def getOrCreateGSIAcquisitionProperty(self,parent,actionsNode,block):
        iecommon.logDebug('Generating GSIAcquisitionProperty for Block: ' + block.prop('name'), {'debuglog': True})
        versionProp = parent.xpathEval('GSI-Version-Property')[0] # template is used --> there has to be a version prop
        prop =  getOrCreateNamedPreviousSiblingElement(parent,versionProp, 'GSI-Acquisition-Property',getFesaName(block))
        fillAttributes(prop, {'visibility': 'development', 'subscribable': 'true', 'multiplexed': 'false', 'on-change': 'true'})
        self.getOrCreateRegisterValueItems(prop,block,False)
        self.getOrCreateAcqStampItem(prop)
        self.getOrCreateUpdateFlagItem(prop)
        self.getOrCreateCyleNameItem(prop)
        self.getOrCreateCyleStampItem(prop)
        self.getOrCreateAction(prop,'Get'+getFesaName(block),'get',actionsNode,'default')
        self.getOrCreateAcquisitionContextItem(prop)
        return prop

    # propertyType only used during creation ! If named property already exists, it is just returned, no matter which type
    def getOrCreateFESAProperty(self,parent,actionsNode,block):
        properties = parent.xpathEval("*")
        if self.isGSITemplate(parent):
            if parent.get_name() == 'setting':
                return self.getOrCreateGSISettingProperty(parent,block)
            else:
                return self.getOrCreateGSIAcquisitionProperty(parent,actionsNode,block)
        else:
            if parent.get_name() == 'setting':
                return self.getOrCreateSettingProperty(parent,block)
            else:
                return self.getOrCreateAcquisitionProperty(parent,actionsNode,block)

    def getOrCreateValueItem(self,propertyNode,register,isSetting):
        iecommon.logDebug('Generating ValueItem for Register: ' + register.prop('name'), {'debuglog': True})
        result = propertyNode.xpathEval("value-item[@name='" + getFesaName(register) + "']")
        if len(result):
            return result[0]
        if( hasChildren(propertyNode)):
            result = propertyNode.xpathEval("*[not(name()='description') and not(name()='export') and not(name()='filter-item')]") # skip possible first elements to get right position
            valueItemNode =  getOrCreateNamedPreviousSiblingElement(propertyNode,result[0],'value-item',getFesaName(register))
        else:
            valueItemNode = getOrCreateNamedChildElement(propertyNode,'value-item',getFesaName(register))
        if isSetting:
            fillAttributes(valueItemNode, {'direction': 'INOUT'})
        else:
            fillAttributes(valueItemNode, {'direction': 'OUT'})
        self.getOrCreateType(valueItemNode,register)
        self.getOrCreateFieldRef(valueItemNode,getFesaName(register))

    #-------------------------------------------------------------------------
    # Generates the data node of the FESA design document
    # parsing the SILECS class document
    #-------------------------------------------------------------------------
    def genData(self,silecsRoot, doc,logTopics):
        iecommon.logDebug("Creating data fields", logTopics)
        
        dataNode = doc.xpathEval('/equipment-model/data')[0]
        iecommon.logError("len(dataNode)" + dataNode.get_name(), logTopics)
        deviceDataNode = getOrCreateChildElement(dataNode,'device-data')
        configurationNode = getOrCreateChildElement(deviceDataNode,'configuration')
        settingNode = getOrCreateChildElement(deviceDataNode,'setting')
        acquisitionNode = getOrCreateChildElement(deviceDataNode,'acquisition')
        self.getOrCreatePLCHostNameField(configurationNode)
        self.getOrCreatePLCDeviceLabelField(configurationNode)
        classNode = silecsRoot.xpathEval('//SILECS-Class')[0]
        self.getOrCreateParameterFileField(configurationNode)
        # parse the SILECS design document to find the blocks 
        for block in silecsRoot.xpathEval("//Block"):   # loop over the blocks
            iecommon.logDebug('Analysing block '+block.prop('name'), logTopics)
            # Check that every register in the block has the same value for synchro
            regList = block.xpathEval("Register")
            for r in range (0, len(regList)-1):
                if regList[r].prop('synchro') != regList[r+1].prop('synchro'):
                    # Return error and exit
                     iecommon.logError('ERROR: In block '+block.prop('name')+' registers have different synchronisation values.', True, logTopics)        
            
            if block.prop('mode') == 'READ-ONLY':
                for reg in regList:
                    fieldNode = self.getOrCreateFieldNode(acquisitionNode,reg)
                    if reg.prop('synchro') == 'MASTER': # the value of a master register never change, consistency is not required
                        fillAttributes(fieldNode, {'multiplexed': 'false', 'persistent': 'false'})
                    elif reg.prop('synchro') == 'NONE':
                        fillAttributes(fieldNode, {'multiplexed': 'false', 'persistent': 'false'})
                    self.getOrCreateType(fieldNode,reg)
            
            elif block.prop('mode') == 'WRITE-ONLY':
                for reg in regList:
                    fieldNode = self.getOrCreateFieldNode(settingNode,reg)
                    if reg.prop('synchro') == 'SLAVE':
                        fillAttributes(fieldNode, {'multiplexed': 'false', 'persistent': 'true', 'shared': 'true'})
                    elif reg.prop('synchro') == 'NONE':
                        fillAttributes(fieldNode, {'multiplexed': 'false', 'persistent': 'false', 'shared': 'true'})
                    self.getOrCreateType(fieldNode,reg)
             
            else:       # READ-WRITE block
                for reg in regList:
                    fieldNode = self.getOrCreateFieldNode(settingNode,reg)
                    if reg.prop('synchro') == 'NONE':
                        fillAttributes(fieldNode, {'multiplexed': 'false', 'persistent': 'false'})
                    else:   #synchro SLAVE and MASTER
                        fillAttributes(fieldNode, {'multiplexed': 'false', 'persistent': 'true'})
                    self.getOrCreateType(fieldNode,reg)

        globalDataNode = getOrCreateChildElement(dataNode,'global-data')
        globalConfigurationNode = ""
        if len(globalDataNode.xpathEval("configuration")) == 0:
            if len( globalDataNode.xpathEval("*")) == 0:
                globalConfigurationNode = getOrCreateChildElement(globalDataNode,'configuration')
            else:
                globalConfigurationNode = getOrCreatePreviousSiblingElement(getFirstChild(globalDataNode),'configuration')
        else:
            globalConfigurationNode = globalDataNode.xpathEval("configuration")[0]
        self.getOrCreatePLCClassVersionField(globalConfigurationNode,classNode.prop('version'))

    #-------------------------------------------------------------------------
    # Generates the interface/device-interface node of the 
    # FESA design document parsing the SILECS class document
    #-------------------------------------------------------------------------
    def genDeviceInterface(self,silecsRoot, doc,logTopics):
        # parse the FESA design document to find the device-interface node and create children
        interfaceNode = doc.xpathEval('/equipment-model/interface')[0]
        globalInterfaceNode = getOrCreateChildElement(interfaceNode,'global-interface')
        deviceInterfaceNode = getOrCreatePreviousSiblingElement(globalInterfaceNode,'device-interface')
        settingNode = getOrCreateChildElement(deviceInterfaceNode,'setting')
        acquisitionNode = getOrCreateChildElement(deviceInterfaceNode,'acquisition')
        actionsNode = doc.xpathEval('/equipment-model/actions')[0]
        for block in silecsRoot.xpathEval("//Block"):

            if block.prop('generateFesaProperty') == 'false':
                continue #skip this block

            if block.prop('mode') in ('READ-ONLY'):
                self.getOrCreateFESAProperty(acquisitionNode,actionsNode,block)
                rtAction = getOrCreateNamedChildElement(actionsNode,'rt-action','Recv'+getFesaName(block))
                notifiedProperty = getOrCreateNamedChildElement(rtAction,'notified-property',getFesaName(block),'property-name-ref')
                fillAttributes(notifiedProperty, {'automatic': 'true'})
                self.genEvent(block,doc,logTopics)
                self.genSchedulingUnit(block,doc,logTopics)
            else:   # setting property for both WO and RW blocks, w/get+set server action
                settingPropertyNode = self.getOrCreateFESAProperty(settingNode, actionsNode, block)
                fillAttributes(settingPropertyNode, {'visibility': 'development', 'multiplexed': 'false'})
                self.getOrCreateAction(settingPropertyNode,'Send'+getFesaName(block),'set',actionsNode,'custom')
                if block.prop('mode') == 'READ-WRITE':
                    self.getOrCreateAction(settingPropertyNode,'Recv'+getFesaName(block),'get',actionsNode,'custom')
                else:   # WRITE-ONLY
                    self.getOrCreateAction(settingPropertyNode,'Get'+getFesaName(block),'get',actionsNode,'default')

    #-------------------------------------------------------------------------
    # Generates the logical-events node of the FESA design
    # document parsing the SILECS class document
    #-------------------------------------------------------------------------       
    def genEvent(self,block, fesaRoot,logTopics):
        iecommon.logDebug('Generating event', logTopics)
        eqpModelNode = fesaRoot.xpathEval("/equipment-model")[0]
        eventsNode = getOrCreateChildElement(eqpModelNode,'events')
        sourcesNode = getOrCreateChildElement(eventsNode,'sources')
        timingSourceNode = getOrCreateNamedChildElement(sourcesNode,'timing-event-source','Timing')
        timerSourceNode = getOrCreateNamedChildElement(sourcesNode,'timer-event-source','Timer')
        logicalEventsNode = getOrCreateChildElement(eventsNode,'logical-events')
        logicalEventNode = getOrCreateNamedChildElement(logicalEventsNode,'logical-event','Recv'+getFesaName(block)+'Event')
        fillAttributes(logicalEventNode, {'type': 'timer', 'use': 'required'})
            
    #-------------------------------------------------------------------------
    # Generates the scheduling-units node of the FESA design
    # document parsing the SILECS class document
    #-------------------------------------------------------------------------
    def genSchedulingUnit(self,block, fesaRoot,logTopics):
        iecommon.logDebug('Generating scheduling unit', logTopics)
        eqpModelNode = fesaRoot.xpathEval("/equipment-model")[0]
        schedulingUnitsNode = getOrCreateChildElement(eqpModelNode,'scheduling-units')
        schedulingUnitNode = getOrCreateNamedChildElement(schedulingUnitsNode,'scheduling-unit','Recv'+getFesaName(block)+'Unit')
        rtActionRefNode = getOrCreateChildElement(schedulingUnitNode,'rt-action-ref')
        fillAttributes(rtActionRefNode, {'rt-action-name-ref': 'Recv'+getFesaName(block)})
        logicalEventRefNode = getOrCreateChildElement(schedulingUnitNode,'logical-event-ref')
        fillAttributes(logicalEventRefNode, {'logical-event-name-ref': 'Recv'+getFesaName(block)+'Event'})

    #-------------------------------------------------------------------------
    # Generates the scheduling-units node of the FESA design
    # document parsing the SILECS class document
    #-------------------------------------------------------------------------
    def addDesignFileId(self,fesaDesignFile,logTopics):
        nodeRequiringId = ['setting-property','acquisition-property','diagnostic-property',
                           'value-item','update-flag-item','cycle-name-item', 'acq-stamp-item',
                           'cycle-stamp-item','mode-item','host-item','port-item','config-item',
                           'state-item','fwk-topic-item','custom-topic-item','device-trace-item',
                           'bypass-action-item','diag-custom-topic','field','timing-event-source',
                           'timer-event-source','logical-event']
        counter = 0
        now = datetime.datetime.today()
        id = '_' + now.strftime('%y%m%d%H%M%S') + '_%s'
        for nodeName in nodeRequiringId:
            for node in fesaDesignFile.xpathEval("//" + nodeName):
                if node.hasProp("id"):
                    node.setProp("id", id%counter)
                    counter = counter +1
        
            
    def fillXML(self,fesaVersion, className, fesaRoot, silecsRoot,logTopics={'errorlog': True}):

        #print etree.tostring(silecsRoot, pretty_print=True)
        #-------------------------------------------------------------------------
        # Fill section <information>
        #-------------------------------------------------------------------------
        informationNode = fesaRoot.xpathEval("/equipment-model/information")[0]
        informationNode.xpathEval("class-name")[0].setContent(className)
        informationNode.xpathEval("fesa-version")[0].setContent(fesaVersion)

        #-------------------------------------------------------------------------
        # Fill section <ownership>
        #-------------------------------------------------------------------------
        creatorNode = fesaRoot.xpathEval("//ownership/creator")[0]
        owner = silecsRoot.xpathEval('//Information/Owner')[0]

        creatorNode.setProp('login', owner.prop('user-login'))    # == Owner in SILECS design document
       
        # Generate data fields
        self.genData(silecsRoot, fesaRoot,logTopics)

        # Generate properties, Actions, Events, Scheduling Units
        self.genDeviceInterface(silecsRoot, fesaRoot,logTopics)

        self.addDesignFileId(fesaRoot,logTopics)
        return fesaRoot

#-------------------------------------------------------------------------
# Fill the FESA design document - entry point of the plugin
#-------------------------------------------------------------------------
def fillDesignFile(fesaVersion, className, workspacePath,fesaDesignXSDPath,logTopics={'errorlog': True}):
    generator = FESADesignGenerator3_0_0()
    fesa.fillFesaDesignBase.fillDesignFileBase(generator,fesaVersion, className, workspacePath,fesaDesignXSDPath,logTopics)

