/*
 * AllTypes.cpp
 *
 * Generated by SILECS framework tools 
 */
 
#include <AllTypes/Common/AllTypes.h>
#include <fesa-core/Synchronization/NoneContext.h>
#include <fesa-core/Synchronization/MultiplexingContext.h>
 
namespace AllTypes
{
    //Global objects of the SILECS class
    Silecs::Service* AbstractAllTypes::pService_ = NULL;
    Silecs::Cluster* AbstractAllTypes::pCluster_ = NULL;
    bool AbstractAllTypes::isInitialized_ = false;

    MyROBlock_Type	AllTypes::MyROBlock("MyROBlock");
    MyRWBlock_Type	AllTypes::MyRWBlock("MyRWBlock");
    MyWOBlock_Type	AllTypes::MyWOBlock("MyWOBlock");
    
    //-------------------------------------------------------------------------------------------------------------
    // Constructor & Destructor methods
    
    AbstractAllTypes::AbstractAllTypes(std::string blockName): blockName_(blockName) {}
    AbstractAllTypes::~AbstractAllTypes() {}
    
    MyROBlock_Type::MyROBlock_Type(std::string name): AbstractAllTypes(name) {}
    MyROBlock_Type::~MyROBlock_Type() {}
    
    MyRWBlock_Type::MyRWBlock_Type(std::string name): AbstractAllTypes(name) {}
    MyRWBlock_Type::~MyRWBlock_Type() {}
    
    MyWOBlock_Type::MyWOBlock_Type(std::string name): AbstractAllTypes(name) {}
    MyWOBlock_Type::~MyWOBlock_Type() {}
    
    //---------------------------------------------------------------------------------------------------------
    // Set-up the SILECS components for the AbstractAllTypes class (service & cluster)

    void AbstractAllTypes::setup(const ServiceLocator* serviceLocator)
    {
        try
        {
            // Instantiate the singleton of the SILECS Service
            pService_ = Silecs::Service::getInstance();

            // Enable the SILECS diagnostic with user topics if any
            pService_->setArguments(serviceLocator->getUsrCmdArgs());

            // Instantiate the SILECS Cluster object for the given Class/Version
            GlobalDevice* pGlobalDevice = serviceLocator->getGlobalDevice();
            pCluster_ = pService_->getCluster( "AllTypes", pGlobalDevice->plcClassVersion.get());
            isInitialized_ = true;

            // Connect each PLC of the Cluster that is referred from the FESA instance
            std::vector<Device*> pDeviceCol = serviceLocator->getDeviceCollection();
            for(std::vector<Device*>::iterator pDeviceIter=pDeviceCol.begin(); pDeviceIter!= pDeviceCol.end(); pDeviceIter++)
            {
                Device* pDevice = *pDeviceIter;

                // Retrieve the PLC related to the current FESA device
                // (from 'plcHostName' FESA field defined on that purpose).
                Silecs::PLC* pPLC = pCluster_->getPLC(pDevice->plcHostName.get(),pDevice->parameterFile.get());

                // Update the PLC Slave registers from related FESA fields just before synchronising done at connection time
                setPLCSlaveRegisters(pPLC, serviceLocator);

                // Connect the PLC if not already connected
                if (!pPLC->isEnabled())
                {   pPLC->connect(/*synchroMode=*/Silecs::FULL_SYNCHRO, /*connectNow=*/true);
                    if (pPLC->isConnected())
                    {   // Update FESA fields from related PLC Master registers just after synchronising done at connection time
                        getPLCMasterRegisters(pPLC, serviceLocator);
                    }
                }
            }
        }
        catch (const Silecs::SilecsException& ex)
        {
            throw fesa::FesaException(__FILE__, __LINE__, ex.getMessage());
        }
    }

    //---------------------------------------------------------------------------------------------------------
    // Release all the SILECS resources
    void AbstractAllTypes::cleanup()
    {
        // Attention! This method is responsible to stop all the PLC connections
        // and to remove all the SILECS resources (Clusters and related components: PLCs, Devices, Registers, ..)
        // Calling method must ensure that no process is currently accessing these resources before cleaning.
        //
        Silecs::Service::deleteInstance();
    }

    //---------------------------------------------------------------------------------------------------------
    // Synchronise PLC SLAVE/MASTER registers and related FESA fields (automatically called by the setup method @connection time)
    void AbstractAllTypes::setPLCSlaveRegisters(Silecs::PLC* pPLC, const ServiceLocator* serviceLocator)
    {
        fesa::NoneContext noneContext;
    
        AllTypes::MyWOBlock.setPLCDevices(pPLC, serviceLocator, false, &noneContext);
    }

    void AbstractAllTypes::getPLCMasterRegisters(Silecs::PLC* pPLC, const ServiceLocator* serviceLocator)
    {
        fesa::NoneContext noneContext;    //MASTER acquisition fields are not consistent, can be set with none-context
    
        AllTypes::MyROBlock.getPLCDevices(pPLC, serviceLocator, false, &noneContext);
        AllTypes::MyRWBlock.getPLCDevices(pPLC, serviceLocator, false, &noneContext);
    }
    
    //---------------------------------------------------------------------------------------------------------
    // General methods to synchronize the FESA fields and related PLC registers of the FESA server
    // with or without PLC side access (send/recv) if requested.
    // get_ : [receive block from PLC +] update FESA fields with related PLC registers
    // set_ : update PLC registers with related FESA fields [+ send block to PLC]

    //---------------------------------------------------------------------------------------------------------

    void AbstractAllTypes::getAllDevices(const ServiceLocator* serviceLocator, const bool recvNow, MultiplexingContext* pContext)
    {
        if (recvNow) theCluster()->recv(blockName_);

        std::vector<Device*> deviceCol = serviceLocator->getDeviceCollection();
        for(std::vector<Device*>::iterator pDeviceIter=deviceCol.begin(); pDeviceIter!= deviceCol.end(); pDeviceIter++)
        {  getOneDevice(*pDeviceIter, false, pContext);
        }
    }

    void AbstractAllTypes::getPLCDevices(Silecs::PLC* pPLC, const ServiceLocator* serviceLocator, const bool recvNow, MultiplexingContext* pContext)
    {
        if (recvNow) pPLC->recv(blockName_);

        std::vector<Device*> deviceCol = serviceLocator->getDeviceCollection();
        for(std::vector<Device*>::iterator pDeviceIter=deviceCol.begin(); pDeviceIter!= deviceCol.end(); pDeviceIter++)
        {   if ((*pDeviceIter)->plcHostName.get() == pPLC->getName())
            {  getOneDevice(*pDeviceIter, false, pContext);
            }
        }
    }

    void AbstractAllTypes::getSomeDevices(std::vector<Device*> deviceCol, const bool sendNow, MultiplexingContext* pContext)
    {
        for(std::vector<Device*>::iterator pDeviceIter=deviceCol.begin(); pDeviceIter!= deviceCol.end(); pDeviceIter++)
        {    getOneDevice(*pDeviceIter, sendNow, pContext);
        }
    }

    void AbstractAllTypes::getOneDevice(Device* pDevice, const bool recvNow, MultiplexingContext* pContext) {};

    //---------------------------------------------------------------------------------------------------------

    void AbstractAllTypes::setAllDevices(const ServiceLocator* serviceLocator, bool sendNow, MultiplexingContext* pContext)
    {
        std::vector<Device*> deviceCol = serviceLocator->getDeviceCollection();
        for(std::vector<Device*>::iterator pDeviceIter=deviceCol.begin(); pDeviceIter!= deviceCol.end(); pDeviceIter++)
        {  setOneDevice(*pDeviceIter, false, pContext);
        }

        if (sendNow) theCluster()->send(blockName_);
    }

    void AbstractAllTypes::setPLCDevices(Silecs::PLC* pPLC, const ServiceLocator* serviceLocator, bool sendNow, MultiplexingContext* pContext)
    {
        std::vector<Device*> deviceCol = serviceLocator->getDeviceCollection();
        for(std::vector<Device*>::iterator pDeviceIter=deviceCol.begin(); pDeviceIter!= deviceCol.end(); pDeviceIter++)
        {   if ((*pDeviceIter)->plcHostName.get() == pPLC->getName())
            {  setOneDevice(*pDeviceIter, false, pContext);
            }
        }

        if (sendNow) pPLC->send(blockName_);
    }

    void AbstractAllTypes::setSomeDevices(std::vector<Device*> deviceCol, bool sendNow, MultiplexingContext* pContext)
    {
        for(std::vector<Device*>::iterator pDeviceIter=deviceCol.begin(); pDeviceIter!= deviceCol.end(); pDeviceIter++)
        {    setOneDevice(*pDeviceIter, sendNow, pContext);
        }
    }

    void AbstractAllTypes::setOneDevice(Device* pDevice, const bool sendNow, MultiplexingContext* pContext) {};
    
    //---------------------------------------------------------------------------------------------------------
    
    void MyROBlock_Type::getOneDevice(Device* pDevice, const bool recvNow, MultiplexingContext* pContext)
    {
    	if( !isInitialized_ )
    	{
    		throw fesa::FesaException(__FILE__, __LINE__, "SILECS-Service not initialized yet - AllTypes::setup needs to be called before any plc-interaction can be done");
    	}
        Silecs::PLC* pPLC  = pCluster_->getPLC(pDevice->plcHostName.get(),pDevice->parameterFile.get());
        Silecs::Device* pPLCDevice = pPLC->getDevice(pDevice->plcDeviceLabel.get());
        Silecs::Register*  pRegister = NULL;

        if (recvNow) pPLCDevice -> recv(blockName_);        
    
        pDevice->RO_int8.set( pPLCDevice->getRegister("RO_int8")->getValInt8(), pContext);
        pDevice->RO_uint8_fesa.set( pPLCDevice->getRegister("RO_uint8")->getValUInt8(), pContext);
        pDevice->RO_int16_fesa.set( pPLCDevice->getRegister("RO_int16")->getValInt16(), pContext);
        pDevice->RO_uint16_fesa.set( pPLCDevice->getRegister("RO_uint16")->getValUInt16(), pContext);
        pDevice->RO_int32_fesa.set( pPLCDevice->getRegister("RO_int32")->getValInt32(), pContext);
        pDevice->RO_uint32_fesa.set( pPLCDevice->getRegister("RO_uint32")->getValUInt32(), pContext);
        pDevice->RO_float32_fesa.set( pPLCDevice->getRegister("RO_float32")->getValFloat32(), pContext);
        pRegister = pPLCDevice->getRegister("RO_string");
        pDevice->RO_string_fesa.set(pRegister->getValString().c_str(), pContext);

        pDevice->RO_date_fesa.set( pPLCDevice->getRegister("RO_date")->getValDate(), pContext);
        pDevice->RO_char_fesa.set( pPLCDevice->getRegister("RO_char")->getValInt8(), pContext);
        pDevice->RO_byte_fesa.set( pPLCDevice->getRegister("RO_byte")->getValUInt8(), pContext);
        pDevice->RO_word_fesa.set( pPLCDevice->getRegister("RO_word")->getValUInt16(), pContext);
        pDevice->RO_dword_fesa.set( pPLCDevice->getRegister("RO_dword")->getValUInt32(), pContext);
        pDevice->RO_int_fesa.set( pPLCDevice->getRegister("RO_int")->getValInt16(), pContext);
        pDevice->RO_dint_fesa.set( pPLCDevice->getRegister("RO_dint")->getValInt32(), pContext);
        pDevice->RO_real_fesa.set( pPLCDevice->getRegister("RO_real")->getValFloat32(), pContext);
        pDevice->RO_dt_fesa.set( pPLCDevice->getRegister("RO_dt")->getValDate(), pContext);
	}
    void MyRWBlock_Type::getOneDevice(Device* pDevice, const bool recvNow, MultiplexingContext* pContext)
    {
    	if( !isInitialized_ )
    	{
    		throw fesa::FesaException(__FILE__, __LINE__, "SILECS-Service not initialized yet - AllTypes::setup needs to be called before any plc-interaction can be done");
    	}
        Silecs::PLC* pPLC  = pCluster_->getPLC(pDevice->plcHostName.get(),pDevice->parameterFile.get());
        Silecs::Device* pPLCDevice = pPLC->getDevice(pDevice->plcDeviceLabel.get());
        Silecs::Register*  pRegister = NULL;
        uint32_t dim1 = 1;
        uint32_t dim2 = 1;

        if (recvNow) pPLCDevice -> recv(blockName_);        
    
        pRegister = pPLCDevice->getRegister("RW_int8");
        dim1 = pRegister->getDimension1();
        dim2 = pRegister->getDimension2();
        pDevice->RW_int8.set(pRegister->getRefInt8Array2D(dim1, dim2), dim1, dim2, pContext);

        pRegister = pPLCDevice->getRegister("RW_uint8");
        dim1 = pRegister->getDimension1();
        dim2 = pRegister->getDimension2();
        int16_t* RW_uint8 = (int16_t*)calloc(dim1*dim2, sizeof(int16_t));
        pRegister->getValUInt8Array2D(RW_uint8, dim1, dim2);	//use automatic conversion for JAVA non-supported type
        pDevice->RW_uint8_fesa.set(RW_uint8, dim1, dim2, pContext);
    free(RW_uint8);

        pRegister = pPLCDevice->getRegister("RW_int16");
        dim1 = pRegister->getDimension1();
        dim2 = pRegister->getDimension2();
        pDevice->RW_int16_fesa.set(pRegister->getRefInt16Array2D(dim1, dim2), dim1, dim2, pContext);

        pRegister = pPLCDevice->getRegister("RW_uint16");
        dim1 = pRegister->getDimension1();
        dim2 = pRegister->getDimension2();
        int32_t* RW_uint16 = (int32_t*)calloc(dim1*dim2, sizeof(int32_t));
        pRegister->getValUInt16Array2D(RW_uint16, dim1, dim2);	//use automatic conversion for JAVA non-supported type
        pDevice->RW_uint16_fesa.set(RW_uint16, dim1, dim2, pContext);
    free(RW_uint16);

        pRegister = pPLCDevice->getRegister("RW_int32");
        dim1 = pRegister->getDimension1();
        dim2 = pRegister->getDimension2();
        pDevice->RW_int32_fesa.set(pRegister->getRefInt32Array2D(dim1, dim2), dim1, dim2, pContext);

        pRegister = pPLCDevice->getRegister("RW_uint32");
        dim1 = pRegister->getDimension1();
        dim2 = pRegister->getDimension2();
        int64_t* RW_uint32 = (int64_t*)calloc(dim1*dim2, sizeof(int64_t));
        pRegister->getValUInt32Array2D(RW_uint32, dim1, dim2);	//use automatic conversion for JAVA non-supported type
        pDevice->RW_uint32_fesa.set(RW_uint32, dim1, dim2, pContext);
    free(RW_uint32);

        pRegister = pPLCDevice->getRegister("RW_float32");
        dim1 = pRegister->getDimension1();
        dim2 = pRegister->getDimension2();
        pDevice->RW_float32_fesa.set(pRegister->getRefFloat32Array2D(dim1, dim2), dim1, dim2, pContext);

        pRegister = pPLCDevice->getRegister("RW_date");
        dim1 = pRegister->getDimension1();
        dim2 = pRegister->getDimension2();
        pDevice->RW_date_fesa.set(pRegister->getRefDateArray2D(dim1, dim2), dim1, dim2, pContext);

        pRegister = pPLCDevice->getRegister("RW_char");
        dim1 = pRegister->getDimension1();
        dim2 = pRegister->getDimension2();
        pDevice->RW_char_fesa.set(pRegister->getRefInt8Array2D(dim1, dim2), dim1, dim2, pContext);

        pRegister = pPLCDevice->getRegister("RW_byte");
        dim1 = pRegister->getDimension1();
        dim2 = pRegister->getDimension2();
        int16_t* RW_byte = (int16_t*)calloc(dim1*dim2, sizeof(int16_t));
        pRegister->getValUInt8Array2D(RW_byte, dim1, dim2);	//use automatic conversion for JAVA non-supported type
        pDevice->RW_byte_fesa.set(RW_byte, dim1, dim2, pContext);
    free(RW_byte);

        pRegister = pPLCDevice->getRegister("RW_word");
        dim1 = pRegister->getDimension1();
        dim2 = pRegister->getDimension2();
        int32_t* RW_word = (int32_t*)calloc(dim1*dim2, sizeof(int32_t));
        pRegister->getValUInt16Array2D(RW_word, dim1, dim2);	//use automatic conversion for JAVA non-supported type
        pDevice->RW_word_fesa.set(RW_word, dim1, dim2, pContext);
    free(RW_word);

        pRegister = pPLCDevice->getRegister("RW_dword");
        dim1 = pRegister->getDimension1();
        dim2 = pRegister->getDimension2();
        int64_t* RW_dword = (int64_t*)calloc(dim1*dim2, sizeof(int64_t));
        pRegister->getValUInt32Array2D(RW_dword, dim1, dim2);	//use automatic conversion for JAVA non-supported type
        pDevice->RW_dword_fesa.set(RW_dword, dim1, dim2, pContext);
    free(RW_dword);

        pRegister = pPLCDevice->getRegister("RW_int");
        dim1 = pRegister->getDimension1();
        dim2 = pRegister->getDimension2();
        pDevice->RW_int_fesa.set(pRegister->getRefInt16Array2D(dim1, dim2), dim1, dim2, pContext);

        pRegister = pPLCDevice->getRegister("RW_dint");
        dim1 = pRegister->getDimension1();
        dim2 = pRegister->getDimension2();
        pDevice->RW_dint_fesa.set(pRegister->getRefInt32Array2D(dim1, dim2), dim1, dim2, pContext);

        pRegister = pPLCDevice->getRegister("RW_real");
        dim1 = pRegister->getDimension1();
        dim2 = pRegister->getDimension2();
        pDevice->RW_real_fesa.set(pRegister->getRefFloat32Array2D(dim1, dim2), dim1, dim2, pContext);

        pRegister = pPLCDevice->getRegister("RW_dt");
        dim1 = pRegister->getDimension1();
        dim2 = pRegister->getDimension2();
        pDevice->RW_dt_fesa.set(pRegister->getRefDateArray2D(dim1, dim2), dim1, dim2, pContext);

	}    
    void MyRWBlock_Type::setOneDevice(Device* pDevice, const bool sendNow, MultiplexingContext* pContext)
    {
    	if( !isInitialized_ )
    	{
    		throw fesa::FesaException(__FILE__, __LINE__, "SILECS-Service not initialized yet - AllTypes::setup needs to be called before any plc-interaction can be done");
    	}
        Silecs::PLC* pPLC  = pCluster_->getPLC(pDevice->plcHostName.get(),pDevice->parameterFile.get());
        Silecs::Device* pPLCDevice = pPLC->getDevice(pDevice->plcDeviceLabel.get());
        Silecs::Register*  pRegister = NULL;
        uint32_t dim1 = 1;
        uint32_t fesaDim1;
        uint32_t dim2 = 1;
        uint32_t fesaDim2;


    pRegister = pPLCDevice->getRegister("RW_int8");
        dim1 = pRegister->getDimension1();
        dim2 = pRegister->getDimension2();
        pRegister->setValInt8Array2D(pDevice->RW_int8.get(fesaDim1, fesaDim2, pContext), dim1, dim2);

        pRegister = pPLCDevice->getRegister("RW_uint8");
        dim1 = pRegister->getDimension1();
        dim2 = pRegister->getDimension2();
        pRegister->setValUInt8Array2D( pDevice->RW_uint8_fesa.get(fesaDim1, fesaDim2, pContext), dim1, dim2);	//use automatic conversion for JAVA non-supported type

    pRegister = pPLCDevice->getRegister("RW_int16");
        dim1 = pRegister->getDimension1();
        dim2 = pRegister->getDimension2();
        pRegister->setValInt16Array2D(pDevice->RW_int16_fesa.get(fesaDim1, fesaDim2, pContext), dim1, dim2);

        pRegister = pPLCDevice->getRegister("RW_uint16");
        dim1 = pRegister->getDimension1();
        dim2 = pRegister->getDimension2();
        pRegister->setValUInt16Array2D( pDevice->RW_uint16_fesa.get(fesaDim1, fesaDim2, pContext), dim1, dim2);	//use automatic conversion for JAVA non-supported type

    pRegister = pPLCDevice->getRegister("RW_int32");
        dim1 = pRegister->getDimension1();
        dim2 = pRegister->getDimension2();
        pRegister->setValInt32Array2D(pDevice->RW_int32_fesa.get(fesaDim1, fesaDim2, pContext), dim1, dim2);

        pRegister = pPLCDevice->getRegister("RW_uint32");
        dim1 = pRegister->getDimension1();
        dim2 = pRegister->getDimension2();
        pRegister->setValUInt32Array2D( pDevice->RW_uint32_fesa.get(fesaDim1, fesaDim2, pContext), dim1, dim2);	//use automatic conversion for JAVA non-supported type

    pRegister = pPLCDevice->getRegister("RW_float32");
        dim1 = pRegister->getDimension1();
        dim2 = pRegister->getDimension2();
        pRegister->setValFloat32Array2D(pDevice->RW_float32_fesa.get(fesaDim1, fesaDim2, pContext), dim1, dim2);

    pRegister = pPLCDevice->getRegister("RW_date");
        dim1 = pRegister->getDimension1();
        dim2 = pRegister->getDimension2();
        pRegister->setValDateArray2D(pDevice->RW_date_fesa.get(fesaDim1, fesaDim2, pContext), dim1, dim2);

    pRegister = pPLCDevice->getRegister("RW_char");
        dim1 = pRegister->getDimension1();
        dim2 = pRegister->getDimension2();
        pRegister->setValInt8Array2D(pDevice->RW_char_fesa.get(fesaDim1, fesaDim2, pContext), dim1, dim2);

        pRegister = pPLCDevice->getRegister("RW_byte");
        dim1 = pRegister->getDimension1();
        dim2 = pRegister->getDimension2();
        pRegister->setValUInt8Array2D( pDevice->RW_byte_fesa.get(fesaDim1, fesaDim2, pContext), dim1, dim2);	//use automatic conversion for JAVA non-supported type

        pRegister = pPLCDevice->getRegister("RW_word");
        dim1 = pRegister->getDimension1();
        dim2 = pRegister->getDimension2();
        pRegister->setValUInt16Array2D( pDevice->RW_word_fesa.get(fesaDim1, fesaDim2, pContext), dim1, dim2);	//use automatic conversion for JAVA non-supported type

        pRegister = pPLCDevice->getRegister("RW_dword");
        dim1 = pRegister->getDimension1();
        dim2 = pRegister->getDimension2();
        pRegister->setValUInt32Array2D( pDevice->RW_dword_fesa.get(fesaDim1, fesaDim2, pContext), dim1, dim2);	//use automatic conversion for JAVA non-supported type

    pRegister = pPLCDevice->getRegister("RW_int");
        dim1 = pRegister->getDimension1();
        dim2 = pRegister->getDimension2();
        pRegister->setValInt16Array2D(pDevice->RW_int_fesa.get(fesaDim1, fesaDim2, pContext), dim1, dim2);

    pRegister = pPLCDevice->getRegister("RW_dint");
        dim1 = pRegister->getDimension1();
        dim2 = pRegister->getDimension2();
        pRegister->setValInt32Array2D(pDevice->RW_dint_fesa.get(fesaDim1, fesaDim2, pContext), dim1, dim2);

    pRegister = pPLCDevice->getRegister("RW_real");
        dim1 = pRegister->getDimension1();
        dim2 = pRegister->getDimension2();
        pRegister->setValFloat32Array2D(pDevice->RW_real_fesa.get(fesaDim1, fesaDim2, pContext), dim1, dim2);

    pRegister = pPLCDevice->getRegister("RW_dt");
        dim1 = pRegister->getDimension1();
        dim2 = pRegister->getDimension2();
        pRegister->setValDateArray2D(pDevice->RW_dt_fesa.get(fesaDim1, fesaDim2, pContext), dim1, dim2);

        if (sendNow) pPLCDevice->send(blockName_);

	}
    void MyRWBlock_Type::setOneDevice(Device* pDevice, MyRWBlockPropPropertyData& data, bool sendNow, MultiplexingContext* pContext)
    {
    	if( !isInitialized_ )
    	{
    		throw fesa::FesaException(__FILE__, __LINE__, "SILECS-Service not initialized yet - AllTypes::setup needs to be called before any plc-interaction can be done");
    	}
        Silecs::PLC* pPLC = pCluster_->getPLC(pDevice->plcHostName.get(),pDevice->parameterFile.get());
        Silecs::Device* pPLCDevice = pPLC->getDevice(pDevice->plcDeviceLabel.get());
        Silecs::Register*  pRegister = NULL;
        uint32_t dim1 = 1;
        uint32_t fesaDim1;
        uint32_t dim2 = 1;
        uint32_t fesaDim2;


        pRegister = pPLCDevice->getRegister("RW_int8");
        dim1 = pRegister->getDimension1();
        dim2 = pRegister->getDimension2();
        (_upper}Available()) ? pRegister->setValInt8Array2D(data.RW_int8.get(fesaDim1, fesaDim2), dim1, dim2) :
                                     pRegister->setValInt8Array2D(pDevice->RW_int8.get(fesaDim1, fesaDim2, pContext), dim1, dim2);

        pRegister = pPLCDevice->getRegister("RW_uint8");
        dim1 = pRegister->getDimension1();
        dim2 = pRegister->getDimension2();
        (data.isRW_uint8_fesaAvailable()) ? pRegister->setValUInt8Array2D(data.RW_uint8_fesa.get(fesaDim1, fesaDim2), dim1, dim2) :
                                     pRegister->setValUInt8Array2D(pDevice->RW_uint8_fesa.get(fesaDim1, fesaDim2, pContext), dim1, dim2);

        pRegister = pPLCDevice->getRegister("RW_int16");
        dim1 = pRegister->getDimension1();
        dim2 = pRegister->getDimension2();
        (_upper}Available()) ? pRegister->setValInt16Array2D(data.RW_int16_fesa.get(fesaDim1, fesaDim2), dim1, dim2) :
                                     pRegister->setValInt16Array2D(pDevice->RW_int16_fesa.get(fesaDim1, fesaDim2, pContext), dim1, dim2);

        pRegister = pPLCDevice->getRegister("RW_uint16");
        dim1 = pRegister->getDimension1();
        dim2 = pRegister->getDimension2();
        (data.isRW_uint16_fesaAvailable()) ? pRegister->setValUInt16Array2D(data.RW_uint16_fesa.get(fesaDim1, fesaDim2), dim1, dim2) :
                                     pRegister->setValUInt16Array2D(pDevice->RW_uint16_fesa.get(fesaDim1, fesaDim2, pContext), dim1, dim2);

        pRegister = pPLCDevice->getRegister("RW_int32");
        dim1 = pRegister->getDimension1();
        dim2 = pRegister->getDimension2();
        (_upper}Available()) ? pRegister->setValInt32Array2D(data.RW_int32_fesa.get(fesaDim1, fesaDim2), dim1, dim2) :
                                     pRegister->setValInt32Array2D(pDevice->RW_int32_fesa.get(fesaDim1, fesaDim2, pContext), dim1, dim2);

        pRegister = pPLCDevice->getRegister("RW_uint32");
        dim1 = pRegister->getDimension1();
        dim2 = pRegister->getDimension2();
        (data.isRW_uint32_fesaAvailable()) ? pRegister->setValUInt32Array2D(data.RW_uint32_fesa.get(fesaDim1, fesaDim2), dim1, dim2) :
                                     pRegister->setValUInt32Array2D(pDevice->RW_uint32_fesa.get(fesaDim1, fesaDim2, pContext), dim1, dim2);

        pRegister = pPLCDevice->getRegister("RW_float32");
        dim1 = pRegister->getDimension1();
        dim2 = pRegister->getDimension2();
        (_upper}Available()) ? pRegister->setValFloat32Array2D(data.RW_float32_fesa.get(fesaDim1, fesaDim2), dim1, dim2) :
                                     pRegister->setValFloat32Array2D(pDevice->RW_float32_fesa.get(fesaDim1, fesaDim2, pContext), dim1, dim2);

        pRegister = pPLCDevice->getRegister("RW_date");
        dim1 = pRegister->getDimension1();
        dim2 = pRegister->getDimension2();
        (_upper}Available()) ? pRegister->setValDateArray2D(data.RW_date_fesa.get(fesaDim1, fesaDim2), dim1, dim2) :
                                     pRegister->setValDateArray2D(pDevice->RW_date_fesa.get(fesaDim1, fesaDim2, pContext), dim1, dim2);

        pRegister = pPLCDevice->getRegister("RW_char");
        dim1 = pRegister->getDimension1();
        dim2 = pRegister->getDimension2();
        (_upper}Available()) ? pRegister->setValInt8Array2D(data.RW_char_fesa.get(fesaDim1, fesaDim2), dim1, dim2) :
                                     pRegister->setValInt8Array2D(pDevice->RW_char_fesa.get(fesaDim1, fesaDim2, pContext), dim1, dim2);

        pRegister = pPLCDevice->getRegister("RW_byte");
        dim1 = pRegister->getDimension1();
        dim2 = pRegister->getDimension2();
        (data.isRW_byte_fesaAvailable()) ? pRegister->setValUInt8Array2D(data.RW_byte_fesa.get(fesaDim1, fesaDim2), dim1, dim2) :
                                     pRegister->setValUInt8Array2D(pDevice->RW_byte_fesa.get(fesaDim1, fesaDim2, pContext), dim1, dim2);

        pRegister = pPLCDevice->getRegister("RW_word");
        dim1 = pRegister->getDimension1();
        dim2 = pRegister->getDimension2();
        (data.isRW_word_fesaAvailable()) ? pRegister->setValUInt16Array2D(data.RW_word_fesa.get(fesaDim1, fesaDim2), dim1, dim2) :
                                     pRegister->setValUInt16Array2D(pDevice->RW_word_fesa.get(fesaDim1, fesaDim2, pContext), dim1, dim2);

        pRegister = pPLCDevice->getRegister("RW_dword");
        dim1 = pRegister->getDimension1();
        dim2 = pRegister->getDimension2();
        (data.isRW_dword_fesaAvailable()) ? pRegister->setValUInt32Array2D(data.RW_dword_fesa.get(fesaDim1, fesaDim2), dim1, dim2) :
                                     pRegister->setValUInt32Array2D(pDevice->RW_dword_fesa.get(fesaDim1, fesaDim2, pContext), dim1, dim2);

        pRegister = pPLCDevice->getRegister("RW_int");
        dim1 = pRegister->getDimension1();
        dim2 = pRegister->getDimension2();
        (_upper}Available()) ? pRegister->setValInt16Array2D(data.RW_int_fesa.get(fesaDim1, fesaDim2), dim1, dim2) :
                                     pRegister->setValInt16Array2D(pDevice->RW_int_fesa.get(fesaDim1, fesaDim2, pContext), dim1, dim2);

        pRegister = pPLCDevice->getRegister("RW_dint");
        dim1 = pRegister->getDimension1();
        dim2 = pRegister->getDimension2();
        (_upper}Available()) ? pRegister->setValInt32Array2D(data.RW_dint_fesa.get(fesaDim1, fesaDim2), dim1, dim2) :
                                     pRegister->setValInt32Array2D(pDevice->RW_dint_fesa.get(fesaDim1, fesaDim2, pContext), dim1, dim2);

        pRegister = pPLCDevice->getRegister("RW_real");
        dim1 = pRegister->getDimension1();
        dim2 = pRegister->getDimension2();
        (_upper}Available()) ? pRegister->setValFloat32Array2D(data.RW_real_fesa.get(fesaDim1, fesaDim2), dim1, dim2) :
                                     pRegister->setValFloat32Array2D(pDevice->RW_real_fesa.get(fesaDim1, fesaDim2, pContext), dim1, dim2);

        pRegister = pPLCDevice->getRegister("RW_dt");
        dim1 = pRegister->getDimension1();
        dim2 = pRegister->getDimension2();
        (_upper}Available()) ? pRegister->setValDateArray2D(data.RW_dt_fesa.get(fesaDim1, fesaDim2), dim1, dim2) :
                                     pRegister->setValDateArray2D(pDevice->RW_dt_fesa.get(fesaDim1, fesaDim2, pContext), dim1, dim2);

        if (sendNow) pPLCDevice->send(blockName_);

	}
    
    void MyWOBlock_Type::setOneDevice(Device* pDevice, const bool sendNow, MultiplexingContext* pContext)
    {
    	if( !isInitialized_ )
    	{
    		throw fesa::FesaException(__FILE__, __LINE__, "SILECS-Service not initialized yet - AllTypes::setup needs to be called before any plc-interaction can be done");
    	}
        Silecs::PLC* pPLC  = pCluster_->getPLC(pDevice->plcHostName.get(),pDevice->parameterFile.get());
        Silecs::Device* pPLCDevice = pPLC->getDevice(pDevice->plcDeviceLabel.get());
        Silecs::Register*  pRegister = NULL;
        uint32_t dim1 = 1;
        uint32_t fesaDim1;


        pRegister = pPLCDevice->getRegister("WO_int8");
            dim1 = pRegister->getDimension1();
            pRegister->setValInt8Array(pDevice->WO_int8.get(fesaDim1, pContext), dim1);
    
        pRegister = pPLCDevice->getRegister("WO_uint8");
        dim1 = pRegister->getDimension1();
        pRegister->setValUInt8Array(pDevice->WO_uint8_fesa.get(fesaDim1, pContext), dim1);	//use automatic conversion for JAVA non-supported type

        pRegister = pPLCDevice->getRegister("WO_int16");
            dim1 = pRegister->getDimension1();
            pRegister->setValInt16Array(pDevice->WO_int16_fesa.get(fesaDim1, pContext), dim1);
    
        pRegister = pPLCDevice->getRegister("WO_uint16");
        dim1 = pRegister->getDimension1();
        pRegister->setValUInt16Array(pDevice->WO_uint16_fesa.get(fesaDim1, pContext), dim1);	//use automatic conversion for JAVA non-supported type

        pRegister = pPLCDevice->getRegister("WO_int32");
            dim1 = pRegister->getDimension1();
            pRegister->setValInt32Array(pDevice->WO_int32_fesa.get(fesaDim1, pContext), dim1);
    
        pRegister = pPLCDevice->getRegister("WO_uint32");
        dim1 = pRegister->getDimension1();
        pRegister->setValUInt32Array(pDevice->WO_uint32_fesa.get(fesaDim1, pContext), dim1);	//use automatic conversion for JAVA non-supported type

        pRegister = pPLCDevice->getRegister("WO_float32");
            dim1 = pRegister->getDimension1();
            pRegister->setValFloat32Array(pDevice->WO_float32_fesa.get(fesaDim1, pContext), dim1);
    
        {
        pRegister = pPLCDevice->getRegister("WO_string");
        dim1 = pRegister->getDimension1();
        std::string stdStringArray[dim1];
        const char** cStringArray = pDevice->WO_string_fesa.get(fesaDim1, pContext);
        for (unsigned int i=0; i<dim1; i++) stdStringArray[i] = (const char*)cStringArray[i];
        pRegister->setValStringArray(stdStringArray, dim1);
        }

        pRegister = pPLCDevice->getRegister("WO_date");
            dim1 = pRegister->getDimension1();
            pRegister->setValDateArray(pDevice->WO_date_fesa.get(fesaDim1, pContext), dim1);
    
        pRegister = pPLCDevice->getRegister("WO_char");
            dim1 = pRegister->getDimension1();
            pRegister->setValInt8Array(pDevice->WO_char_fesa.get(fesaDim1, pContext), dim1);
    
        pRegister = pPLCDevice->getRegister("WO_byte");
        dim1 = pRegister->getDimension1();
        pRegister->setValUInt8Array(pDevice->WO_byte_fesa.get(fesaDim1, pContext), dim1);	//use automatic conversion for JAVA non-supported type

        pRegister = pPLCDevice->getRegister("WO_word");
        dim1 = pRegister->getDimension1();
        pRegister->setValUInt16Array(pDevice->WO_word_fesa.get(fesaDim1, pContext), dim1);	//use automatic conversion for JAVA non-supported type

        pRegister = pPLCDevice->getRegister("WO_dword");
        dim1 = pRegister->getDimension1();
        pRegister->setValUInt32Array(pDevice->WO_dword_fesa.get(fesaDim1, pContext), dim1);	//use automatic conversion for JAVA non-supported type

        pRegister = pPLCDevice->getRegister("WO_int");
            dim1 = pRegister->getDimension1();
            pRegister->setValInt16Array(pDevice->WO_int_fesa.get(fesaDim1, pContext), dim1);
    
        pRegister = pPLCDevice->getRegister("WO_dint");
            dim1 = pRegister->getDimension1();
            pRegister->setValInt32Array(pDevice->WO_dint_fesa.get(fesaDim1, pContext), dim1);
    
        pRegister = pPLCDevice->getRegister("WO_real");
            dim1 = pRegister->getDimension1();
            pRegister->setValFloat32Array(pDevice->WO_real_fesa.get(fesaDim1, pContext), dim1);
    
        pRegister = pPLCDevice->getRegister("WO_dt");
            dim1 = pRegister->getDimension1();
            pRegister->setValDateArray(pDevice->WO_dt_fesa.get(fesaDim1, pContext), dim1);
    
        if (sendNow) pPLCDevice->send(blockName_);

	}
    void MyWOBlock_Type::setOneDevice(Device* pDevice, MyWOBlockPropPropertyData& data, bool sendNow, MultiplexingContext* pContext)
    {
    	if( !isInitialized_ )
    	{
    		throw fesa::FesaException(__FILE__, __LINE__, "SILECS-Service not initialized yet - AllTypes::setup needs to be called before any plc-interaction can be done");
    	}
        Silecs::PLC* pPLC = pCluster_->getPLC(pDevice->plcHostName.get(),pDevice->parameterFile.get());
        Silecs::Device* pPLCDevice = pPLC->getDevice(pDevice->plcDeviceLabel.get());
        Silecs::Register*  pRegister = NULL;
        uint32_t dim1 = 1;
        uint32_t fesaDim1;


        pRegister = pPLCDevice->getRegister("WO_int8");
            dim1 = pRegister->getDimension1();
            (data.isWO_int8Available()) ? pRegister->setValInt8Array( data.WO_int8.get(fesaDim1), dim1) :
                                         pRegister->setValInt8Array( pDevice->WO_int8.get(fesaDim1, pContext), dim1);
    
        pRegister = pPLCDevice->getRegister("WO_uint8");
            dim1 = pRegister->getDimension1();
            (data.isWO_uint8_fesaAvailable()) ? pRegister->setValUInt8Array( data.WO_uint8_fesa.get(fesaDim1), dim1) :
                                         pRegister->setValUInt8Array( pDevice->WO_uint8_fesa.get(fesaDim1, pContext), dim1);
    
        pRegister = pPLCDevice->getRegister("WO_int16");
            dim1 = pRegister->getDimension1();
            (data.isWO_int16_fesaAvailable()) ? pRegister->setValInt16Array( data.WO_int16_fesa.get(fesaDim1), dim1) :
                                         pRegister->setValInt16Array( pDevice->WO_int16_fesa.get(fesaDim1, pContext), dim1);
    
        pRegister = pPLCDevice->getRegister("WO_uint16");
            dim1 = pRegister->getDimension1();
            (data.isWO_uint16_fesaAvailable()) ? pRegister->setValUInt16Array( data.WO_uint16_fesa.get(fesaDim1), dim1) :
                                         pRegister->setValUInt16Array( pDevice->WO_uint16_fesa.get(fesaDim1, pContext), dim1);
    
        pRegister = pPLCDevice->getRegister("WO_int32");
            dim1 = pRegister->getDimension1();
            (data.isWO_int32_fesaAvailable()) ? pRegister->setValInt32Array( data.WO_int32_fesa.get(fesaDim1), dim1) :
                                         pRegister->setValInt32Array( pDevice->WO_int32_fesa.get(fesaDim1, pContext), dim1);
    
        pRegister = pPLCDevice->getRegister("WO_uint32");
            dim1 = pRegister->getDimension1();
            (data.isWO_uint32_fesaAvailable()) ? pRegister->setValUInt32Array( data.WO_uint32_fesa.get(fesaDim1), dim1) :
                                         pRegister->setValUInt32Array( pDevice->WO_uint32_fesa.get(fesaDim1, pContext), dim1);
    
        pRegister = pPLCDevice->getRegister("WO_float32");
            dim1 = pRegister->getDimension1();
            (data.isWO_float32_fesaAvailable()) ? pRegister->setValFloat32Array( data.WO_float32_fesa.get(fesaDim1), dim1) :
                                         pRegister->setValFloat32Array( pDevice->WO_float32_fesa.get(fesaDim1, pContext), dim1);
    
        {
        pRegister = pPLCDevice->getRegister("WO_string");
        dim1 = pRegister->getDimension1();
        std::string stdStringArray[dim1];
        const char** cStringArray = (data.isWO_string_fesaAvailable() ? data.WO_string_fesa.get(fesaDim1) : pDevice->WO_string_fesa.get(fesaDim1, pContext));
        for (unsigned int i=0; i<dim1; i++) stdStringArray[i] = (const char*)cStringArray[i];
        pRegister->setValStringArray(stdStringArray, dim1);
        }

        pRegister = pPLCDevice->getRegister("WO_date");
            dim1 = pRegister->getDimension1();
            (data.isWO_date_fesaAvailable()) ? pRegister->setValDateArray( data.WO_date_fesa.get(fesaDim1), dim1) :
                                         pRegister->setValDateArray( pDevice->WO_date_fesa.get(fesaDim1, pContext), dim1);
    
        pRegister = pPLCDevice->getRegister("WO_char");
            dim1 = pRegister->getDimension1();
            (data.isWO_char_fesaAvailable()) ? pRegister->setValInt8Array( data.WO_char_fesa.get(fesaDim1), dim1) :
                                         pRegister->setValInt8Array( pDevice->WO_char_fesa.get(fesaDim1, pContext), dim1);
    
        pRegister = pPLCDevice->getRegister("WO_byte");
            dim1 = pRegister->getDimension1();
            (data.isWO_byte_fesaAvailable()) ? pRegister->setValUInt8Array( data.WO_byte_fesa.get(fesaDim1), dim1) :
                                         pRegister->setValUInt8Array( pDevice->WO_byte_fesa.get(fesaDim1, pContext), dim1);
    
        pRegister = pPLCDevice->getRegister("WO_word");
            dim1 = pRegister->getDimension1();
            (data.isWO_word_fesaAvailable()) ? pRegister->setValUInt16Array( data.WO_word_fesa.get(fesaDim1), dim1) :
                                         pRegister->setValUInt16Array( pDevice->WO_word_fesa.get(fesaDim1, pContext), dim1);
    
        pRegister = pPLCDevice->getRegister("WO_dword");
            dim1 = pRegister->getDimension1();
            (data.isWO_dword_fesaAvailable()) ? pRegister->setValUInt32Array( data.WO_dword_fesa.get(fesaDim1), dim1) :
                                         pRegister->setValUInt32Array( pDevice->WO_dword_fesa.get(fesaDim1, pContext), dim1);
    
        pRegister = pPLCDevice->getRegister("WO_int");
            dim1 = pRegister->getDimension1();
            (data.isWO_int_fesaAvailable()) ? pRegister->setValInt16Array( data.WO_int_fesa.get(fesaDim1), dim1) :
                                         pRegister->setValInt16Array( pDevice->WO_int_fesa.get(fesaDim1, pContext), dim1);
    
        pRegister = pPLCDevice->getRegister("WO_dint");
            dim1 = pRegister->getDimension1();
            (data.isWO_dint_fesaAvailable()) ? pRegister->setValInt32Array( data.WO_dint_fesa.get(fesaDim1), dim1) :
                                         pRegister->setValInt32Array( pDevice->WO_dint_fesa.get(fesaDim1, pContext), dim1);
    
        pRegister = pPLCDevice->getRegister("WO_real");
            dim1 = pRegister->getDimension1();
            (data.isWO_real_fesaAvailable()) ? pRegister->setValFloat32Array( data.WO_real_fesa.get(fesaDim1), dim1) :
                                         pRegister->setValFloat32Array( pDevice->WO_real_fesa.get(fesaDim1, pContext), dim1);
    
        pRegister = pPLCDevice->getRegister("WO_dt");
            dim1 = pRegister->getDimension1();
            (data.isWO_dt_fesaAvailable()) ? pRegister->setValDateArray( data.WO_dt_fesa.get(fesaDim1), dim1) :
                                         pRegister->setValDateArray( pDevice->WO_dt_fesa.get(fesaDim1, pContext), dim1);
    
        if (sendNow) pPLCDevice->send(blockName_);

	}
}
